@@ -198,8 +198,10 @@ static void convert_bitmaps(struct gfs2_sbd *sdp, struct rgrp_tree *rg)
 			for (y = 0; y < GFS2_NBBY; y++) {
 				state = (rg->bh[blk]->b_data[x] >>
 					 (GFS2_BIT_SIZE * y)) & 0x03;
-				if (state == 0x02) /* unallocated metadata state invalid */
+				if (state == 0x02) {/* unallocated metadata state invalid */
 					rg->bh[blk]->b_data[x] &= ~(0x02 << (GFS2_BIT_SIZE * y));
+					bmodified(rg->bh[blk]);
+				}
 			}
 	}
 }/* convert_bitmaps */
@@ -235,7 +237,7 @@ static int convert_rgs(struct gfs2_sbd *sbp)
 		sbp->dinodes_alloced += rgd1->rg_useddi;
 		convert_bitmaps(sbp, rgd);
 		/* Write the updated rgrp to the gfs2 buffer */
-		gfs2_rgrp_out(&rgd->rg, rgd->bh[0]);
+		gfs2_rgrp_out_bh(&rgd->rg, rgd->bh[0]);
 		rgs++;
 		if (rgs % 100 == 0) {
 			printf(".");
@@ -270,6 +272,12 @@ static unsigned int calc_gfs2_tree_height(struct gfs2_inode *ip, uint64_t size)
 	for (height = 0; height < max; height++)
 		if (arr[height] >= size)
 			break;
+	/* If calc_gfs2_tree_height was called, the dinode is not stuffed or
+	   we would have returned before this point. After the call, a call is
+	   made to fix_metatree, which unstuffs the dinode. Therefore, the
+	   smallest height that can result after this call is 1. */
+	if (!height)
+		height = 1;
 
 	return height;
 }
@@ -351,10 +359,11 @@ static void fix_metatree(struct gfs2_sbd *sbp, struct gfs2_inode *ip,
 			bh = bread(sbp, block);
 			if (new)
 				memset(bh->b_data, 0, sbp->bsize);
-			gfs2_meta_header_out(&mh, bh);
+			gfs2_meta_header_out_bh(&mh, bh);
 		}
 
-		hdrsize = sizeof(struct gfs2_meta_header);
+		hdrsize = blk->height ? sizeof(struct gfs2_meta_header) :
+			sizeof(struct gfs2_dinode);
 
 		if (amount > sbp->bsize - hdrsize - ptramt)
 			amount = sbp->bsize - hdrsize - ptramt;
@@ -508,7 +517,7 @@ static void fix_jdatatree(struct gfs2_sbd *sbp, struct gfs2_inode *ip,
 			if (new)
 				memset(bh->b_data, 0, sbp->bsize);
 			if (h < (blk->height - 1))
-				gfs2_meta_header_out(&mh, bh);
+				gfs2_meta_header_out_bh(&mh, bh);
 		}
 
 		if (amount > sbp->bsize - ptramt)
@@ -821,7 +830,6 @@ static int has_cdpn(const char *str)
 
 static int fix_cdpn_symlink(struct gfs2_sbd *sbp, struct gfs2_buffer_head *bh, struct gfs2_inode *ip)
 {
-	int ret = 0;
 	char *linkptr = NULL;
 
 	if (ip->i_di.di_height != 0)
@@ -842,7 +850,7 @@ static int fix_cdpn_symlink(struct gfs2_sbd *sbp, struct gfs2_buffer_head *bh, s
 				  (osi_list_t *)&cdpns_to_fix);
 	}
 
-	return ret;
+	return 0;
 }
 
 /*
@@ -857,7 +865,7 @@ static int fix_cdpn_symlink(struct gfs2_sbd *sbp, struct gfs2_buffer_head *bh, s
  */
 static int fix_xattr(struct gfs2_sbd *sbp, struct gfs2_buffer_head *bh, struct gfs2_inode *ip)
 {
-	int ret = 0, len, old_hdr_sz, new_hdr_sz;
+	int len, old_hdr_sz, new_hdr_sz;
 	struct gfs2_buffer_head *eabh;
 	char *buf;
 
@@ -881,7 +889,7 @@ static int fix_xattr(struct gfs2_sbd *sbp, struct gfs2_buffer_head *bh, struct g
 	}
         brelse(eabh);
 
-	return ret;
+	return 0;
 }
 
 /* ------------------------------------------------------------------------- */
@@ -895,14 +903,19 @@ static int adjust_inode(struct gfs2_sbd *sbp, struct gfs2_buffer_head *bh)
 	struct inode_block *fixdir;
 	int inode_was_gfs1;
 
-	inode = gfs_inode_get(sbp, bh);
+	inode = lgfs2_gfs_inode_get(sbp, bh);
+	if (inode == NULL) {
+		log_crit(_("Error reading inode: %s\n"), strerror(errno));
+		return -1;
+	}
 
 	inode_was_gfs1 = (inode->i_di.di_num.no_formal_ino ==
 					  inode->i_di.di_num.no_addr);
 	/* Fix the inode number: */
-	inode->i_di.di_num.no_formal_ino = sbp->md.next_inum;           ;
-	
+	inode->i_di.di_num.no_formal_ino = sbp->md.next_inum;
+
 	/* Fix the inode type: gfs1 uses di_type, gfs2 uses di_mode. */
+	inode->i_di.di_mode &= ~S_IFMT;
 	switch (inode->i_di.__pad1) { /* formerly di_type */
 	case GFS_FILE_DIR:           /* directory        */
 		inode->i_di.di_mode |= S_IFDIR;
@@ -967,7 +980,7 @@ static int adjust_inode(struct gfs2_sbd *sbp, struct gfs2_buffer_head *bh)
 		if (adjust_indirect_blocks(sbp, inode))
 			return -1;
 		/* Check for cdpns */
-		if (inode->i_di.di_mode & S_IFLNK) {
+		if (S_ISLNK(inode->i_di.di_mode)) {
 			ret = fix_cdpn_symlink(sbp, bh, inode);
 			if (ret)
 				return -1;
@@ -986,6 +999,56 @@ static int adjust_inode(struct gfs2_sbd *sbp, struct gfs2_buffer_head *bh)
 	return 0;
 } /* adjust_inode */
 
+static int next_rg_meta(struct rgrp_tree *rgd, uint64_t *block, int first)
+{
+	struct gfs2_bitmap *bits = NULL;
+	uint32_t length = rgd->ri.ri_length;
+	uint32_t blk = (first)? 0: (uint32_t)((*block + 1) - rgd->ri.ri_data0);
+	int i;
+
+	if (!first && (*block < rgd->ri.ri_data0)) {
+		fprintf(stderr, "next_rg_meta:  Start block is outside rgrp bounds.\n");
+		exit(1);
+	}
+	for (i = 0; i < length; i++){
+		bits = &rgd->bits[i];
+		if (blk < bits->bi_len * GFS2_NBBY)
+			break;
+		blk -= bits->bi_len * GFS2_NBBY;
+	}
+	for (; i < length; i++){
+		bits = &rgd->bits[i];
+		blk = gfs2_bitfit((unsigned char *)rgd->bh[i]->b_data +
+				  bits->bi_offset, bits->bi_len, blk, GFS2_BLKST_DINODE);
+		if(blk != BFITNOENT){
+			*block = blk + (bits->bi_start * GFS2_NBBY) +
+				rgd->ri.ri_data0;
+			break;
+		}
+		blk = 0;
+	}
+	if (i == length)
+		return -1;
+	return 0;
+}
+
+static int next_rg_metatype(struct gfs2_sbd *sdp, struct rgrp_tree *rgd,
+                            uint64_t *block, uint32_t type, int first)
+{
+	struct gfs2_buffer_head *bh = NULL;
+
+	do{
+		if (bh)
+			brelse(bh);
+		if (next_rg_meta(rgd, block, first))
+			return -1;
+		bh = bread(sdp, *block);
+		first = 0;
+	} while(gfs2_check_meta(bh, type));
+	brelse(bh);
+	return 0;
+}
+
 /* ------------------------------------------------------------------------- */
 /* inode_renumber - renumber the inodes                                      */
 /*                                                                           */
@@ -998,7 +1061,7 @@ static int inode_renumber(struct gfs2_sbd *sbp, uint64_t root_inode_addr, osi_li
 {
 	struct rgrp_tree *rgd;
 	struct osi_node *n, *next = NULL;
-	uint64_t block;
+	uint64_t block = 0;
 	struct gfs2_buffer_head *bh;
 	int first;
 	int error = 0;
@@ -1033,7 +1096,7 @@ static int inode_renumber(struct gfs2_sbd *sbp, uint64_t root_inode_addr, osi_li
 			/* be "11" (used meta) for both inodes and indirect blocks.     */
 			/* We need to process the inodes and change the indirect blocks */
 			/* to have a bitmap type of "01" (data).                        */
-			if (gfs2_next_rg_metatype(sbp, rgd, &block, 0, first))
+			if (next_rg_metatype(sbp, rgd, &block, 0, first))
 				break;
 			/* If this is the root inode block, remember it for later: */
 			if (block == root_inode_addr) {
@@ -1068,10 +1131,10 @@ static int inode_renumber(struct gfs2_sbd *sbp, uint64_t root_inode_addr, osi_li
 							~(0x03 << (GFS2_BIT_SIZE * byte_bit));
 						rgd->bh[blk]->b_data[buf_offset + bitmap_byte] |=
 							(0x01 << (GFS2_BIT_SIZE * byte_bit));
+						bmodified(rgd->bh[blk]);
 						break;
 					}
 					bitmap_byte -= (sbp->bsize - buf_offset);
-					bmodified(rgd->bh[blk]);
 				}
 			}
 			brelse(bh);
@@ -1092,7 +1155,9 @@ static int fetch_inum(struct gfs2_sbd *sbp, uint64_t iblock,
 {
 	struct gfs2_inode *fix_inode;
 
-	fix_inode = inode_read(sbp, iblock);
+	fix_inode = lgfs2_inode_read(sbp, iblock);
+	if (fix_inode == NULL)
+		return 1;
 	inum->no_formal_ino = fix_inode->i_di.di_num.no_formal_ino;
 	inum->no_addr = fix_inode->i_di.di_num.no_addr;
 	if (eablk)
@@ -1289,7 +1354,9 @@ static int process_directory(struct gfs2_sbd *sbp, uint64_t dirblock, uint64_t d
 	struct gfs2_inode *dip;
 	int error = 0;
 	/* read in the directory inode */
-	dip = inode_read(sbp, dirblock);
+	dip = lgfs2_inode_read(sbp, dirblock);
+	if (dip == NULL)
+		return -1;
 	/* fix the directory: either exhash (leaves) or linear (stuffed) */
 	if (dip->i_di.di_flags & GFS2_DIF_EXHASH) {
 		if (fix_one_directory_exhash(sbp, dip, dentmod)) {
@@ -1391,7 +1458,9 @@ static int fix_cdpn_symlinks(struct gfs2_sbd *sbp, osi_list_t *cdpn_to_fix)
 
 		/* initialize the symlink inode to be a directory */
 		bh = init_dinode(sbp, &fix, S_IFDIR | 0755, 0, &dir);
-		fix_inode = inode_get(sbp, bh);
+		fix_inode = lgfs2_inode_get(sbp, bh);
+		if (fix_inode == NULL)
+			return -1;
 		fix_inode->i_di.di_eattr = eablk; /*fix extended attribute */
 		inode_put(&fix_inode);
 		bmodified(bh);
@@ -1585,10 +1654,18 @@ static int init(struct gfs2_sbd *sbp)
 	}
 	/* get gfs1 rindex inode - gfs1's rindex inode ptr became __pad2 */
 	gfs2_inum_in(&inum, (char *)&raw_gfs1_ondisk_sb.sb_rindex_di);
-	sbp->md.riinode = gfs_inode_read(sbp, inum.no_addr);
+	sbp->md.riinode = lgfs2_gfs_inode_read(sbp, inum.no_addr);
+	if (sbp->md.riinode == NULL) {
+		log_crit(_("Could not read resource group index: %s\n"), strerror(errno));
+		exit(-1);
+	}
 	/* get gfs1 jindex inode - gfs1's journal index inode ptr became master */
 	gfs2_inum_in(&inum, (char *)&raw_gfs1_ondisk_sb.sb_jindex_di);
-	sbp->md.jiinode = inode_read(sbp, inum.no_addr);
+	sbp->md.jiinode = lgfs2_inode_read(sbp, inum.no_addr);
+	if (sbp->md.jiinode == NULL) {
+		log_crit(_("Could not read journal index: %s\n"), strerror(errno));
+		exit(-1);
+	}
 	/* read in the journal index data */
 	read_gfs1_jiindex(sbp);
 	/* read in the resource group index data: */
@@ -1777,13 +1854,13 @@ static int journ_space_to_rg(struct gfs2_sbd *sdp)
 				 (rgd->ri.ri_addr > rgdhigh->ri.ri_addr)))
 				rgdhigh = rgd;
 		} /* for each rg */
-		log_info(_("Addr 0x%llx comes after rg at addr 0x%llx\n"),
-			 (unsigned long long)jndx->ji_addr,
-			 (unsigned long long)rgdhigh->ri.ri_addr);
 		if (!rgdhigh) { /* if we somehow didn't find one. */
 			log_crit(_("Error: No suitable rg found for journal.\n"));
 			return -1;
 		}
+		log_info(_("Addr 0x%llx comes after rg at addr 0x%llx\n"),
+			 (unsigned long long)jndx->ji_addr,
+			 (unsigned long long)rgdhigh->ri.ri_addr);
 		ri_addr = jndx->ji_addr;
 		/* Allocate a new rgd entry which includes rg and ri. */
 		rgd = rgrp_insert(&sdp->rgtree, ri_addr);
@@ -1821,16 +1898,16 @@ static int journ_space_to_rg(struct gfs2_sbd *sdp)
 			rgd->bh[x] = bget(sdp, rgd->ri.ri_addr + x);
 			memset(rgd->bh[x]->b_data, 0, sdp->bsize);
 		}
-		if (gfs2_compute_bitstructs(sdp, rgd)) {
+		if (gfs2_compute_bitstructs(sdp->sd_sb.sb_bsize, rgd)) {
 			log_crit(_("gfs2_convert: Error converting bitmaps.\n"));
 			exit(-1);
 		}
 		convert_bitmaps(sdp, rgd);
 		for (x = 0; x < rgd->ri.ri_length; x++) {
 			if (x)
-				gfs2_meta_header_out(&mh, rgd->bh[x]);
+				gfs2_meta_header_out_bh(&mh, rgd->bh[x]);
 			else
-				gfs2_rgrp_out(&rgd->rg, rgd->bh[x]);
+				gfs2_rgrp_out_bh(&rgd->rg, rgd->bh[x]);
 		}
 	} /* for each journal */
 	return error;
@@ -1847,8 +1924,10 @@ static void update_inode_file(struct gfs2_sbd *sdp)
 	
 	buf = cpu_to_be64(sdp->md.next_inum);
 	count = gfs2_writei(ip, &buf, 0, sizeof(uint64_t));
-	if (count != sizeof(uint64_t))
-		die("update_inode_file\n");
+	if (count != sizeof(uint64_t)) {
+		fprintf(stderr, "update_inode_file\n");
+		exit(1);
+	}
 	
 	log_debug(_("\nNext Inum: %llu\n"), (unsigned long long)sdp->md.next_inum);
 }/* update_inode_file */
@@ -1869,8 +1948,10 @@ static void write_statfs_file(struct gfs2_sbd *sdp)
 
 	gfs2_statfs_change_out(&sc, buf);
 	count = gfs2_writei(ip, buf, 0, sizeof(struct gfs2_statfs_change));
-	if (count != sizeof(struct gfs2_statfs_change))
-		die("do_init (2)\n");
+	if (count != sizeof(struct gfs2_statfs_change)) {
+		fprintf(stderr, "do_init (2)\n");
+		exit(1);
+	}
 }/* write_statfs_file */
 
 /* ------------------------------------------------------------------------- */
@@ -1914,6 +1995,9 @@ static int conv_build_jindex(struct gfs2_sbd *sdp)
 
 	sdp->md.journal = malloc(sdp->md.journals *
 				 sizeof(struct gfs2_inode *));
+	if (sdp->md.journal == NULL) {
+		return errno;
+	}
 	for (j = 0; j < sdp->md.journals; j++) {
 		char name[256];
 
@@ -2043,14 +2127,21 @@ static void copy_quotas(struct gfs2_sbd *sdp)
 	int err;
 
 	err = gfs2_lookupi(sdp->master_dir, "quota", 5, &nq_ip);
-	if (err)
-		die(_("Couldn't lookup new quota file: %d\n"), err);
+	if (err) {
+		fprintf(stderr, _("Couldn't lookup new quota file: %d\n"), err);
+		exit(1);
+	}
 
 	gfs2_inum_in(&inum, (char *)&raw_gfs1_ondisk_sb.sb_quota_di);
-	oq_ip = inode_read(sdp, inum.no_addr);
+	oq_ip = lgfs2_inode_read(sdp, inum.no_addr);
+	if (oq_ip == NULL) {
+		fprintf(stderr, _("Couldn't lookup old quota file: %s\n"), strerror(errno));
+		exit(1);
+	}
 
 	nq_ip->i_di.di_height = oq_ip->i_di.di_height;
 	nq_ip->i_di.di_size = oq_ip->i_di.di_size;
+	nq_ip->i_di.di_blocks = oq_ip->i_di.di_blocks;
 	memcpy(nq_ip->i_bh->b_data + sizeof(struct gfs2_dinode), 
 	       oq_ip->i_bh->b_data + sizeof(struct gfs2_dinode),
 	       sdp->bsize - sizeof(struct gfs2_dinode));
@@ -2065,89 +2156,11 @@ static void copy_quotas(struct gfs2_sbd *sdp)
 	inode_put(&oq_ip);
 }
 
-static char gfs2_getch(void)
+static int gfs2_query(struct gfs2_options *opts, const char *dev)
 {
-	struct termios termattr, savetermattr;
-	char ch;
-	ssize_t size;
-
-	tcgetattr (STDIN_FILENO, &termattr);
-	savetermattr = termattr;
-	termattr.c_lflag &= ~(ICANON | IEXTEN | ISIG);
-	termattr.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
-	termattr.c_cflag &= ~(CSIZE | PARENB);
-	termattr.c_cflag |= CS8;
-	termattr.c_oflag &= ~(OPOST);
-   	termattr.c_cc[VMIN] = 0;
-	termattr.c_cc[VTIME] = 0;
-
-	tcsetattr (STDIN_FILENO, TCSANOW, &termattr);
-	do {
-		size = read(STDIN_FILENO, &ch, 1);
-		if (size)
-			break;
-		usleep(50000);
-	} while (!size);
-
-	tcsetattr (STDIN_FILENO, TCSANOW, &savetermattr);
-	return ch;
-}
-
-static char generic_interrupt(const char *caller, const char *where,
-		       const char *progress, const char *question,
-		       const char *answers)
-{
-	fd_set rfds;
-	struct timeval stv;
-	char response;
-	int err, i;
-
-	FD_ZERO(&rfds);
-	FD_SET(STDIN_FILENO, &rfds);
-
-	stv.tv_sec = 0;
-	stv.tv_usec = 0;
-	/* Make sure there isn't extraneous input before asking the
-	 * user the question */
-	while((err = select(STDIN_FILENO + 1, &rfds, NULL, NULL, &stv))) {
-		if(err < 0) {
-			log_debug("Error in select() on stdin\n");
-			break;
-		}
-		if(read(STDIN_FILENO, &response, sizeof(char)) < 0) {
-			log_debug("Error in read() on stdin\n");
-			break;
-		}
-	}
-	while (TRUE) {
-		printf("\n%s interrupted during %s:  ", caller, where);
-		if (progress)
-			printf("%s.\n", progress);
-		printf("%s", question);
-
-		/* Make sure query is printed out */
-		fflush(NULL);
-		response = gfs2_getch();
-		printf("\n");
-		fflush(NULL);
-		if (strchr(answers, response))
-			break;
-		printf("Bad response, please type ");
-		for (i = 0; i < strlen(answers) - 1; i++)
-			printf("'%c', ", answers[i]);
-		printf(" or '%c'.\n", answers[i]);
-	}
-	return response;
-}
-
-static int __attribute__((format(printf, 3, 4))) gfs2_query(int *setonabort,
-                           struct gfs2_options *opts, const char *format, ...)
-{
-	va_list args;
-	char response;
+	char response[3] = { 0, 0 };
 	int ret = 0;
 
-	*setonabort = 0;
 	if(opts->yes)
 		return 1;
 	if(opts->no)
@@ -2155,41 +2168,18 @@ static int __attribute__((format(printf, 3, 4))) gfs2_query(int *setonabort,
 
 	opts->query = TRUE;
 	while (1) {
-		va_start(args, format);
-		vprintf(format, args);
-		va_end(args);
-
+		printf(_("Convert %s from GFS1 to GFS2? (y/n)"), dev);
 		/* Make sure query is printed out */
 		fflush(NULL);
-		response = gfs2_getch();
-
+		fgets(response, 3, stdin);
 		printf("\n");
 		fflush(NULL);
-		if (response == 0x3) { /* if interrupted, by ctrl-c */
+		response[1] = 0;
+		ret = rpmatch(response);
 
-			/*This is ok to translate if nobody changes the (a/c) option.
- 			 * Should we proceed with this translation */
-			response = generic_interrupt(_("Question"), _("response"),
-						     NULL,
-						     _("Do you want to abort " \
-						     "or continue (a/c)?"), 
-						     "ac");
-			if (response == 'a') {
-				ret = 0;
-				*setonabort = 1;
-				break;
-			}
-			printf(_("Continuing.\n"));
-		} else if(tolower(response) == 'y') {
-			ret = 1;
+		if (ret >= 0)
 			break;
-		} else if (tolower(response) == 'n') {
-			ret = 0;
-			break;
-		} else {
-			printf(_("Bad response %d, please type 'y' or 'n'.\n"),
-			       response);
-		}
+		printf(_("Bad response '%s', please type 'y' or 'n'.\n"), response);
 	}
 
 	opts->query = FALSE;
@@ -2217,12 +2207,8 @@ int main(int argc, char **argv)
 	/* Make them seal their fate.                     */
 	/* ---------------------------------------------- */
 	if (!error) {
-		int do_abort;
-
 		give_warning();
-		if (!gfs2_query(&do_abort, &opts,
-				_("Convert %s from GFS1 to GFS2? (y/n)"),
-				device)) {
+		if (!gfs2_query(&opts, device)) {
 			log_crit(_("%s not converted.\n"), device);
 			close(sb2.device_fd);
 			exit(0);
