@@ -30,9 +30,6 @@
 static void print_block_details(struct iinfo *ind, int level, int cur_height,
 				int pndx, uint64_t file_offset);
 
-/* ------------------------------------------------------------------------ */
-/* get_height                                                               */
-/* ------------------------------------------------------------------------ */
 static int get_height(void)
 {
 	int cur_height = 0, i;
@@ -49,9 +46,6 @@ static int get_height(void)
 	return cur_height;
 }
 
-/* ------------------------------------------------------------------------ */
-/* _do_indirect_extended                                                    */
-/* ------------------------------------------------------------------------ */
 static int _do_indirect_extended(char *diebuf, struct iinfo *iinf, int hgt)
 {
 	unsigned int x, y;
@@ -81,9 +75,6 @@ static int _do_indirect_extended(char *diebuf, struct iinfo *iinf, int hgt)
 	return i_blocks;
 }
 
-/* ------------------------------------------------------------------------ */
-/* do_indirect_extended                                                     */
-/* ------------------------------------------------------------------------ */
 int do_indirect_extended(char *diebuf, struct iinfo *iinf)
 {
 	return _do_indirect_extended(diebuf, iinf, get_height());
@@ -107,9 +98,6 @@ static int dinode_valid(void)
 	return 0;
 }
 
-/* ------------------------------------------------------------------------ */
-/* metapath_to_lblock - convert from metapath, height to logical block      */
-/* ------------------------------------------------------------------------ */
 static uint64_t metapath_to_lblock(struct metapath *mp, int hgt)
 {
 	int h;
@@ -128,9 +116,6 @@ static uint64_t metapath_to_lblock(struct metapath *mp, int hgt)
 	return lblock;
 }
 
-/* ------------------------------------------------------------------------ */
-/* display_indirect                                                         */
-/* ------------------------------------------------------------------------ */
 static int display_indirect(struct iinfo *ind, int indblocks, int level,
 			    uint64_t startoff)
 {
@@ -246,9 +231,6 @@ static int display_indirect(struct iinfo *ind, int indblocks, int level,
 	return 0;
 }
 
-/* ------------------------------------------------------------------------ */
-/* print_inode_type                                                         */
-/* ------------------------------------------------------------------------ */
 static void print_inode_type(__be16 de_type)
 {
 	if (sbd.gfs1) {
@@ -314,9 +296,6 @@ static void print_inode_type(__be16 de_type)
 	}
 }
 
-/* ------------------------------------------------------------------------ */
-/* display_leaf - display directory leaf                                    */
-/* ------------------------------------------------------------------------ */
 static int display_leaf(struct iinfo *ind)
 {
 	int start_line, total_dirents = start_row[dmode];
@@ -340,7 +319,7 @@ static int display_leaf(struct iinfo *ind)
 			break;
 		total_dirents++;
 		if (ind->ii[0].dirents >= 1) {
-			eol(5);
+			eol(3);
 			if (termlines) {
 				if (edit_row[dmode] >=0 &&
 				    line - start_line - 1 ==
@@ -351,9 +330,12 @@ static int display_leaf(struct iinfo *ind)
 					strcpy(edit_fmt, "%llx");
 				}
 			}
-			print_gfs2("%d. (%d). %lld (0x%llx): ",
+			print_gfs2("%d/%d [%08x] %lld/%lld (0x%llx/0x%llx): ",
 				   total_dirents, d + 1,
+				   ind->ii[0].dirent[d].dirent.de_hash,
+				   ind->ii[0].dirent[d].dirent.de_inum.no_formal_ino,
 				   ind->ii[0].dirent[d].block,
+				   ind->ii[0].dirent[d].dirent.de_inum.no_formal_ino,
 				   ind->ii[0].dirent[d].block);
 		}
 		print_inode_type(ind->ii[0].dirent[d].dirent.de_type);
@@ -374,9 +356,6 @@ static int display_leaf(struct iinfo *ind)
 	return 0;
 }
 
-/* ------------------------------------------------------------------------ */
-/* print_block_details                                                      */
-/* ------------------------------------------------------------------------ */
 static void print_block_details(struct iinfo *ind, int level, int cur_height,
 				int pndx, uint64_t file_offset)
 {
@@ -396,6 +375,7 @@ static void print_block_details(struct iinfo *ind, int level, int cur_height,
 	if (!tmpbuf) {
 		fprintf(stderr, "Out of memory in function "
 			"display_indirect\n");
+		free(more_indir);
 		return;
 	}
 	while (thisblk) {
@@ -440,9 +420,6 @@ static void print_block_details(struct iinfo *ind, int level, int cur_height,
 	free(more_indir);
 }
 
-/* ------------------------------------------------------------------------ */
-/* gfs_jindex_print - print an jindex entry.                                */
-/* ------------------------------------------------------------------------ */
 static void gfs_jindex_print(struct gfs_jindex *ji)
 {
         pv((unsigned long long)ji, ji_addr, "%llu", "0x%llx");
@@ -450,9 +427,6 @@ static void gfs_jindex_print(struct gfs_jindex *ji)
         pv(ji, ji_pad, "%u", "0x%x");
 }
 
-/* ------------------------------------------------------------------------ */
-/* print_jindex - print the jindex file.                                    */
-/* ------------------------------------------------------------------------ */
 static int print_jindex(struct gfs2_inode *dij)
 {
 	int error, start_line;
@@ -493,14 +467,11 @@ static int print_jindex(struct gfs2_inode *dij)
 	return error;
 }
 
-/* ------------------------------------------------------------------------ */
-/* parse_rindex - print the rgindex file.                                   */
-/* ------------------------------------------------------------------------ */
 static int parse_rindex(struct gfs2_inode *dip, int print_rindex)
 {
 	int error, start_line;
 	struct gfs2_rindex ri;
-	char rbuf[sizeof(struct gfs_rindex)];
+	char rbuf[sizeof(struct gfs2_rindex)];
 	char highlighted_addr[32];
 
 	start_line = line;
@@ -561,9 +532,6 @@ static int parse_rindex(struct gfs2_inode *dip, int print_rindex)
 	return error;
 }
 
-/* ------------------------------------------------------------------------ */
-/* print_inum - print the inum file.                                        */
-/* ------------------------------------------------------------------------ */
 static int print_inum(struct gfs2_inode *dii)
 {
 	uint64_t inum, inodenum;
@@ -586,9 +554,6 @@ static int print_inum(struct gfs2_inode *dii)
 	return 0;
 }
 
-/* ------------------------------------------------------------------------ */
-/* print_statfs - print the statfs file.                                    */
-/* ------------------------------------------------------------------------ */
 static int print_statfs(struct gfs2_inode *dis)
 {
 	struct gfs2_statfs_change sfb, sfc;
@@ -612,9 +577,6 @@ static int print_statfs(struct gfs2_inode *dis)
 	return 0;
 }
 
-/* ------------------------------------------------------------------------ */
-/* print_quota - print the quota file.                                      */
-/* ------------------------------------------------------------------------ */
 static int print_quota(struct gfs2_inode *diq)
 {
 	struct gfs2_quota qbuf, q;
@@ -641,9 +603,6 @@ static int print_quota(struct gfs2_inode *diq)
 	return 0;
 }
 
-/* ------------------------------------------------------------------------ */
-/* display_extended                                                         */
-/* ------------------------------------------------------------------------ */
 int display_extended(void)
 {
 	struct gfs2_inode *tmp_inode;
@@ -653,8 +612,11 @@ int display_extended(void)
 	/* Display any indirect pointers that we have. */
 	if (block_is_rindex()) {
 		tmp_bh = bread(&sbd, block);
-		tmp_inode = inode_get(&sbd, tmp_bh);
+		tmp_inode = lgfs2_inode_get(&sbd, tmp_bh);
+		if (tmp_inode == NULL)
+			return -1;
 		parse_rindex(tmp_inode, TRUE);
+		inode_put(&tmp_inode);
 		brelse(tmp_bh);
 	}
 	else if (has_indirect_blocks() && !indirect_blocks &&
@@ -667,32 +629,47 @@ int display_extended(void)
 			tmp_bh = bread(&sbd, sbd1->sb_rindex_di.no_addr);
 		else
 			tmp_bh = bread(&sbd, masterblock("rindex"));
-		tmp_inode = inode_get(&sbd, tmp_bh);
+		tmp_inode = lgfs2_inode_get(&sbd, tmp_bh);
+		if (tmp_inode == NULL)
+			return -1;
 		parse_rindex(tmp_inode, FALSE);
+		inode_put(&tmp_inode);
 		brelse(tmp_bh);
 	}
 	else if (block_is_jindex()) {
 		tmp_bh = bread(&sbd, block);
-		tmp_inode = inode_get(&sbd, tmp_bh);
+		tmp_inode = lgfs2_inode_get(&sbd, tmp_bh);
+		if (tmp_inode == NULL)
+			return -1;
 		print_jindex(tmp_inode);
+		inode_put(&tmp_inode);
 		brelse(tmp_bh);
 	}
 	else if (block_is_inum_file()) {
 		tmp_bh = bread(&sbd, block);
-		tmp_inode = inode_get(&sbd, tmp_bh);
+		tmp_inode = lgfs2_inode_get(&sbd, tmp_bh);
+		if (tmp_inode == NULL)
+			return -1;
 		print_inum(tmp_inode);
+		inode_put(&tmp_inode);
 		brelse(tmp_bh);
 	}
 	else if (block_is_statfs_file()) {
 		tmp_bh = bread(&sbd, block);
-		tmp_inode = inode_get(&sbd, tmp_bh);
+		tmp_inode = lgfs2_inode_get(&sbd, tmp_bh);
+		if (tmp_inode == NULL)
+			return -1;
 		print_statfs(tmp_inode);
+		inode_put(&tmp_inode);
 		brelse(tmp_bh);
 	}
 	else if (block_is_quota_file()) {
 		tmp_bh = bread(&sbd, block);
-		tmp_inode = inode_get(&sbd, tmp_bh);
+		tmp_inode = lgfs2_inode_get(&sbd, tmp_bh);
+		if (tmp_inode == NULL)
+			return -1;
 		print_quota(tmp_inode);
+		inode_put(&tmp_inode);
 		brelse(tmp_bh);
 	}
 	return 0;
