@@ -43,7 +43,7 @@ struct metafd {
 };
 
 struct saved_metablock *savedata;
-uint64_t last_fs_block, last_reported_block, blks_saved, total_out, pct;
+uint64_t last_reported_block, blks_saved, total_out, pct;
 uint64_t journal_blocks[MAX_JOURNALS_SAVED];
 uint64_t gfs1_journal_size = 0; /* in blocks */
 int journals_found = 0;
@@ -60,6 +60,30 @@ static int block_is_a_journal(void)
 	return FALSE;
 }
 
+static int block_is_in_per_node(void)
+{
+	int d;
+	struct gfs2_inode *per_node_di;
+
+	if (sbd.gfs1)
+		return FALSE;
+
+	per_node_di = lgfs2_inode_read(&sbd, masterblock("per_node"));
+	if (per_node_di == NULL) {
+		fprintf(stderr, "Failed to read per_node: %s\n", strerror(errno));
+		exit(1);
+	}
+
+	do_dinode_extended(&per_node_di->i_di, per_node_di->i_bh);
+	inode_put(&per_node_di);
+
+	for (d = 0; d < indirect->ii[0].dirents; d++) {
+		if (block == indirect->ii[0].dirent[d].block)
+			return TRUE;
+	}
+	return FALSE;
+}
+
 static int block_is_systemfile(void)
 {
 	return block_is_jindex() || block_is_inum_file() ||
@@ -122,10 +146,15 @@ static int get_gfs_struct_info(struct gfs2_buffer_head *lbh, int *block_type,
 		*gstruct_len = sbd.bsize;
 		break;
 	case GFS2_METATYPE_DI:   /* 4 (disk inode) */
-		if (sbd.gfs1)
-			inode = gfs_inode_get(&sbd, lbh);
-		else
-			inode = inode_get(&sbd, lbh);
+		if (sbd.gfs1) {
+			inode = lgfs2_gfs_inode_get(&sbd, lbh);
+		} else {
+			inode = lgfs2_inode_get(&sbd, lbh);
+		}
+		if (inode == NULL) {
+			perror("Error reading inode");
+			exit(-1);
+		}
 		if (S_ISDIR(inode->i_di.di_mode) ||
 		     (sbd.gfs1 && inode->i_di.__pad1 == GFS_FILE_DIR))
 			*gstruct_len = sbd.bsize;
@@ -189,11 +218,13 @@ static void warm_fuzzy_stuff(uint64_t wfsblock, int force)
 		static uint64_t percent;
 
 		seconds = tv.tv_sec;
-		if (last_fs_block) {
+		if (sbd.fssize) {
 			printf("\r");
-			percent = (wfsblock * 100) / last_fs_block;
-			printf("%llu metadata blocks (%llu%%) processed, ",
+			percent = (wfsblock * 100) / sbd.fssize;
+			printf("%llu inodes processed, %llu blocks saved "
+			       "(%llu%%) processed, ",
 			       (unsigned long long)wfsblock,
+			       (unsigned long long)blks_saved,
 			       (unsigned long long)percent);
 			if (force)
 				printf("\n");
@@ -299,7 +330,7 @@ static int save_block(int fd, struct metafd *mfd, uint64_t blk)
 	char *p;
 	struct gfs2_buffer_head *savebh;
 
-	if (blk > last_fs_block) {
+	if (gfs2_check_range(&sbd, blk) && blk != sbd.sb_addr) {
 		fprintf(stderr, "\nWarning: bad block pointer '0x%llx' "
 			"ignored in block (block %llu (0x%llx))",
 			(unsigned long long)blk,
@@ -419,7 +450,8 @@ static void save_indirect_blocks(struct metafd *mfd, osi_list_t *cur_list,
 			save_ea_block(mfd, nbh);
 			brelse(nbh);
 		}
-		if (height != hgt) { /* If not at max height */
+		if (height != hgt && /* If not at max height and */
+		    (!gfs2_check_range(&sbd, indir_block))) {
 			nbh = bread(&sbd, indir_block);
 			osi_list_add_prev(&nbh->b_altlist, cur_list);
 			/* The buffer_head needs to be queued ahead, so
@@ -456,10 +488,15 @@ static void save_inode_data(struct metafd *mfd)
 	for (i = 0; i < GFS2_MAX_META_HEIGHT; i++)
 		osi_list_init(&metalist[i]);
 	metabh = bread(&sbd, block);
-	if (sbd.gfs1)
-		inode = gfs_inode_get(&sbd, metabh);
-	else
-		inode = inode_get(&sbd, metabh);
+	if (sbd.gfs1) {
+		inode = lgfs2_gfs_inode_get(&sbd, metabh);
+	} else {
+		inode = lgfs2_inode_get(&sbd, metabh);
+	}
+	if (inode == NULL) {
+		perror("Failed to read inode");
+		exit(-1);
+	}
 	height = inode->i_di.di_height;
 	/* If this is a user inode, we don't follow to the file height.
 	   We stop one level less.  That way we save off the indirect
@@ -481,6 +518,7 @@ static void save_inode_data(struct metafd *mfd)
 		for (tmp = prev_list->next; tmp != prev_list; tmp = tmp->next){
 			mybh = osi_list_entry(tmp, struct gfs2_buffer_head,
 					      b_altlist);
+			warm_fuzzy_stuff(block, FALSE);
 			save_indirect_blocks(mfd, cur_list, mybh,
 					     height, i);
 		} /* for blocks at that height */
@@ -511,6 +549,7 @@ static void save_inode_data(struct metafd *mfd)
 				continue;
 			old_leaf = leaf_no;
 			mybh = bread(&sbd, leaf_no);
+			warm_fuzzy_stuff(block, FALSE);
 			if (gfs2_check_meta(mybh, GFS2_METATYPE_LF) == 0)
 				save_block(sbd.device_fd, mfd, leaf_no);
 			brelse(mybh);
@@ -572,8 +611,12 @@ static void get_journal_inode_blocks(void)
 			struct gfs_jindex ji;
 			char jbuf[sizeof(struct gfs_jindex)];
 
-			j_inode = gfs_inode_read(&sbd,
+			j_inode = lgfs2_gfs_inode_read(&sbd,
 						 sbd1->sb_jindex_di.no_addr);
+			if (j_inode == NULL) {
+				fprintf(stderr, "Error reading journal inode: %s\n", strerror(errno));
+				return;
+			}
 			amt = gfs2_readi(j_inode, (void *)&jbuf,
 					 journal * sizeof(struct gfs_jindex),
 					 sizeof(struct gfs_jindex));
@@ -592,55 +635,45 @@ static void get_journal_inode_blocks(void)
 	}
 }
 
-static int next_rg_freemeta(struct gfs2_sbd *sdp, struct rgrp_tree *rgd,
-			    uint64_t *nrfblock, int first)
+static void save_allocated(struct rgrp_tree *rgd, struct metafd *mfd)
 {
-	struct gfs2_bitmap *bits = NULL;
-	uint32_t length = rgd->ri.ri_length;
-	uint32_t blk = (first)? 0: (uint32_t)((*nrfblock+1)-rgd->ri.ri_data0);
-	int i;
-	struct gfs2_buffer_head *lbh;
+	int blktype;
+	unsigned i, j, m;
+	uint64_t *ibuf = malloc(sbd.bsize * GFS2_NBBY * sizeof(uint64_t));
 
-	if(!first && (*nrfblock < rgd->ri.ri_data0)) {
-		log_err("next_rg_freemeta:  Start block is outside rgrp "
-			"bounds.\n");
-		exit(1);
-	}
-	for(i=0; i < length; i++){
-		bits = &rgd->bits[i];
-		if(blk < bits->bi_len*GFS2_NBBY)
-			break;
-		blk -= bits->bi_len*GFS2_NBBY;
-	}
-	for(; i < length; i++){
-		bits = &rgd->bits[i];
-		lbh = bread(sdp, rgd->ri.ri_addr + i);
-		blk = gfs2_bitfit((unsigned char *)lbh->b_data +
-				  bits->bi_offset, bits->bi_len, blk,
-				  GFS2_BLKST_UNLINKED);
-		brelse(lbh);
-		if(blk != BFITNOENT){
-			*nrfblock = blk + (bits->bi_start * GFS2_NBBY) +
-				rgd->ri.ri_data0;
-			break;
+	for (i = 0; i < rgd->ri.ri_length; i++) {
+		m = lgfs2_bm_scan(rgd, i, ibuf, GFS2_BLKST_DINODE);
+
+		for (j = 0; j < m; j++) {
+			block = ibuf[j];
+			warm_fuzzy_stuff(block, FALSE);
+			blktype = save_block(sbd.device_fd, mfd, block);
+			if (blktype == GFS2_METATYPE_DI)
+				save_inode_data(mfd);
+		}
+
+		if (!sbd.gfs1)
+			continue;
+
+		/* For gfs1, Save off the free/unlinked meta blocks too.
+		 * If we don't, we may run into metadata allocation issues. */
+		m = lgfs2_bm_scan(rgd, i, ibuf, GFS2_BLKST_UNLINKED);
+		for (j = 0; j < m; j++) {
+			blktype = save_block(sbd.device_fd, mfd, block);
 		}
-		blk=0;
 	}
-	if(i == length)
-		return -1;
-	return 0;
+	free(ibuf);
 }
 
 void savemeta(char *out_fn, int saveoption, int gziplevel)
 {
-	int slow, ret;
 	int rgcount;
 	uint64_t jindex_block;
 	struct gfs2_buffer_head *lbh;
-	struct rgrp_tree *last_rgd;
 	struct metafd mfd;
+	int sane;
+	struct osi_node *n, *next = NULL;
 
-	slow = (saveoption == 1);
 	sbd.md.journals = 1;
 
 	mfd = savemetaopen(out_fn, gziplevel);
@@ -653,150 +686,108 @@ void savemeta(char *out_fn, int saveoption, int gziplevel)
 	blks_saved = total_out = last_reported_block = 0;
 	if (!sbd.gfs1)
 		sbd.bsize = BUFSIZE;
-	if (!slow) {
-		if (lgfs2_get_dev_info(sbd.device_fd, &sbd.dinfo)) {
-			perror(sbd.device_name);
-			exit(-1);
-		}
-		fix_device_geometry(&sbd);
-		sbd.rgtree.osi_node = NULL;
-		if (!sbd.gfs1)
-			sbd.sd_sb.sb_bsize = GFS2_DEFAULT_BSIZE;
-		if (compute_constants(&sbd)) {
-			fprintf(stderr, "Bad constants (1)\n");
-			exit(-1);
-		}
-		ret = read_sb(&sbd);
-		if (ret < 0) {
-			slow = TRUE;
-			sbd.gfs1 = 0;
-		}
-		if (sbd.gfs1)
-			sbd.bsize = sbd.sd_sb.sb_bsize;
+	if (lgfs2_get_dev_info(sbd.device_fd, &sbd.dinfo)) {
+		perror(sbd.device_name);
+		exit(-1);
 	}
-	last_fs_block = lseek(sbd.device_fd, 0, SEEK_END) / sbd.bsize;
+	fix_device_geometry(&sbd);
+	sbd.rgtree.osi_node = NULL;
+	if (!sbd.gfs1)
+		sbd.sd_sb.sb_bsize = GFS2_DEFAULT_BSIZE;
+	if (compute_constants(&sbd)) {
+		fprintf(stderr, "Bad constants (1)\n");
+		exit(-1);
+	}
+	if (read_sb(&sbd) < 0) {
+		fprintf(stderr, "Unable to read superblock.\n");
+		exit(-1);
+	}
+	if (sbd.gfs1)
+		sbd.bsize = sbd.sd_sb.sb_bsize;
 	printf("There are %llu blocks of %u bytes in the destination "
-	       "device.\n", (unsigned long long)last_fs_block, sbd.bsize);
-	if (!slow) {
-		if (sbd.gfs1) {
-			sbd.md.riinode = inode_read(&sbd,
-						sbd1->sb_rindex_di.no_addr);
-			jindex_block = sbd1->sb_jindex_di.no_addr;
-		} else {
-			sbd.master_dir =
-				inode_read(&sbd,
-					sbd.sd_sb.sb_master_dir.no_addr);
-
-			slow = gfs2_lookupi(sbd.master_dir, "rindex", 6, 
-					    &sbd.md.riinode);
-			jindex_block = masterblock("jindex");
+	       "device.\n", (unsigned long long)sbd.fssize, sbd.bsize);
+	if (sbd.gfs1) {
+		sbd.md.riinode = lgfs2_inode_read(&sbd, sbd1->sb_rindex_di.no_addr);
+		if (sbd.md.riinode == NULL) {
+			fprintf(stderr, "Unable to read rindex: %s.\n", strerror(errno));
+			exit(-1);
+		}
+		jindex_block = sbd1->sb_jindex_di.no_addr;
+	} else {
+		sbd.master_dir = lgfs2_inode_read(&sbd,
+					    sbd.sd_sb.sb_master_dir.no_addr);
+		if (sbd.master_dir == NULL) {
+			fprintf(stderr, "Unable to read master: %s.\n", strerror(errno));
+			exit(-1);
 		}
-		lbh = bread(&sbd, jindex_block);
-		gfs2_dinode_in(&di, lbh);
-		if (!sbd.gfs1)
-			do_dinode_extended(&di, lbh);
-		brelse(lbh);
-	}
-	if (!slow) {
-		int sane;
-		uint64_t fssize;
-		struct osi_node *n;
 
-		printf("Reading resource groups...");
-		fflush(stdout);
-		if (sbd.gfs1)
-			slow = gfs1_ri_update(&sbd, 0, &rgcount, 0);
-		else
-			slow = ri_update(&sbd, 0, &rgcount, &sane);
-		n = osi_last(&sbd.rgtree);
-		last_rgd = (struct rgrp_tree *)n;
-		n = osi_prev(n);
-		fssize = last_rgd->ri.ri_addr + rgrp_size(last_rgd);
-		last_fs_block = fssize;
-		fssize *= sbd.bsize;
-		printf("Done. File system size: %s\n\n",
-		       anthropomorphize(fssize));
-		fflush(stdout);
+		gfs2_lookupi(sbd.master_dir, "rindex", 6, &sbd.md.riinode);
+		jindex_block = masterblock("jindex");
 	}
+	lbh = bread(&sbd, jindex_block);
+	gfs2_dinode_in(&di, lbh);
+	if (!sbd.gfs1)
+		do_dinode_extended(&di, lbh);
+	brelse(lbh);
+
+	printf("Reading resource groups...");
+	fflush(stdout);
+	if (sbd.gfs1)
+		gfs1_ri_update(&sbd, 0, &rgcount, 0);
+	else
+		ri_update(&sbd, 0, &rgcount, &sane);
+	printf("Done. File system size: %s\n\n", anthropomorphize(sbd.fssize));
+	fflush(stdout);
+
 	get_journal_inode_blocks();
-	if (!slow) {
-		struct osi_node *n, *next = NULL;
 
-		/* Save off the superblock */
-		save_block(sbd.device_fd, &mfd, 0x10 * (4096 / sbd.bsize));
-		/* If this is gfs1, save off the rindex because it's not
-		   part of the file system as it is in gfs2. */
-		if (sbd.gfs1) {
-			int j;
+	/* Save off the superblock */
+	save_block(sbd.device_fd, &mfd, 0x10 * (4096 / sbd.bsize));
+	/* If this is gfs1, save off the rindex because it's not
+	   part of the file system as it is in gfs2. */
+	if (sbd.gfs1) {
+		int j;
 
-			block = sbd1->sb_rindex_di.no_addr;
-			save_block(sbd.device_fd, &mfd, block);
-			save_inode_data(&mfd);
-			/* In GFS1, journals aren't part of the RG space */
-			for (j = 0; j < journals_found; j++) {
-				log_debug("Saving journal #%d\n", j + 1);
-				for (block = journal_blocks[j];
-				     block < journal_blocks[j] +
-					     gfs1_journal_size;
-				     block++)
-					save_block(sbd.device_fd, &mfd, block);
-			}
-		}
-		/* Walk through the resource groups saving everything within */
-		for (n = osi_first(&sbd.rgtree); n; n = next) {
-			int first;
-			struct rgrp_tree *rgd;
-
-			next = osi_next(n);
-			rgd = (struct rgrp_tree *)n;
-			slow = gfs2_rgrp_read(&sbd, rgd);
-			if (slow)
-				continue;
-			log_debug("RG at %lld (0x%llx) is %u long\n",
-				  (unsigned long long)rgd->ri.ri_addr,
-				  (unsigned long long)rgd->ri.ri_addr,
-				  rgd->ri.ri_length);
-			first = 1;
-			/* Save off the rg and bitmaps */
-			for (block = rgd->ri.ri_addr;
-			     block < rgd->ri.ri_data0; block++) {
-				warm_fuzzy_stuff(block, FALSE);
+		block = sbd1->sb_rindex_di.no_addr;
+		save_block(sbd.device_fd, &mfd, block);
+		save_inode_data(&mfd);
+		/* In GFS1, journals aren't part of the RG space */
+		for (j = 0; j < journals_found; j++) {
+			log_debug("Saving journal #%d\n", j + 1);
+			for (block = journal_blocks[j];
+			     block < journal_blocks[j] + gfs1_journal_size;
+			     block++)
 				save_block(sbd.device_fd, &mfd, block);
-			}
-			/* Save off the other metadata: inodes, etc. */
-			if (saveoption != 2) {
-				int blktype;
-
-				while (!gfs2_next_rg_meta(rgd, &block, first)){
-					warm_fuzzy_stuff(block, FALSE);
-					blktype = save_block(sbd.device_fd,
-							     &mfd, block);
-					if (blktype == GFS2_METATYPE_DI)
-						save_inode_data(&mfd);
-					first = 0;
-				}
-				/* Save off the free/unlinked meta blocks too.
-				   If we don't, we may run into metadata
-				   allocation issues. */
-				while (!next_rg_freemeta(&sbd, rgd, &block,
-							 first)) {
-					blktype = save_block(sbd.device_fd,
-							     &mfd, block);
-					first = 0;
-				}
-			}
-			gfs2_rgrp_relse(rgd);
 		}
 	}
-	if (slow) {
-		for (block = 0; block < last_fs_block; block++) {
+	/* Walk through the resource groups saving everything within */
+	for (n = osi_first(&sbd.rgtree); n; n = next) {
+		struct rgrp_tree *rgd;
+
+		next = osi_next(n);
+		rgd = (struct rgrp_tree *)n;
+		if (gfs2_rgrp_read(&sbd, rgd))
+			continue;
+		log_debug("RG at %lld (0x%llx) is %u long\n",
+			  (unsigned long long)rgd->ri.ri_addr,
+			  (unsigned long long)rgd->ri.ri_addr,
+			  rgd->ri.ri_length);
+		/* Save off the rg and bitmaps */
+		for (block = rgd->ri.ri_addr;
+		     block < rgd->ri.ri_data0; block++) {
+			warm_fuzzy_stuff(block, FALSE);
 			save_block(sbd.device_fd, &mfd, block);
 		}
+		/* Save off the other metadata: inodes, etc. if mode is not 'savergs' */
+		if (saveoption != 2) {
+			save_allocated(rgd, &mfd);
+		}
+		gfs2_rgrp_relse(rgd);
 	}
 	/* Clean up */
 	/* There may be a gap between end of file system and end of device */
 	/* so we tell the user that we've processed everything. */
-	block = last_fs_block;
+	block = sbd.fssize;
 	warm_fuzzy_stuff(block, TRUE);
 	printf("\nMetadata saved to file %s ", mfd.filename);
 	if (mfd.gziplevel) {
@@ -812,7 +803,7 @@ void savemeta(char *out_fn, int saveoption, int gziplevel)
 	exit(0);
 }
 
-static int restore_data(int fd, gzFile *gzin_fd, int printblocksonly,
+static int restore_data(int fd, gzFile gzin_fd, int printblocksonly,
 			int find_highblk)
 {
 	size_t rs;
@@ -848,7 +839,6 @@ static int restore_data(int fd, gzFile *gzin_fd, int printblocksonly,
 		exit(-1);
 	}
 	blks_saved = total_out = 0;
-	last_fs_block = 0;
 	while (TRUE) {
 		struct gfs2_buffer_head dummy_bh;
 
@@ -863,11 +853,11 @@ static int restore_data(int fd, gzFile *gzin_fd, int printblocksonly,
 		total_out += sbd.bsize;
 		savedata->blk = be64_to_cpu(buf64);
 		if (!printblocksonly &&
-		    last_fs_block && savedata->blk >= last_fs_block) {
+		    sbd.fssize && savedata->blk >= sbd.fssize) {
 			fprintf(stderr, "Error: File system is too small to "
 				"restore this metadata.\n");
 			fprintf(stderr, "File system is %llu blocks, ",
-				(unsigned long long)last_fs_block);
+				(unsigned long long)sbd.fssize);
 			fprintf(stderr, "Restore block = %llu\n",
 				(unsigned long long)savedata->blk);
 			return -1;
@@ -918,11 +908,11 @@ static int restore_data(int fd, gzFile *gzin_fd, int printblocksonly,
 			if (find_highblk)
 				;
 			else if (!printblocksonly) {
-				last_fs_block =
+				sbd.fssize =
 					lseek(fd, 0, SEEK_END) / sbd.bsize;
 				printf("There are %llu blocks of %u bytes in "
 				       "the destination device.\n\n",
-				       (unsigned long long)last_fs_block, sbd.bsize);
+				       (unsigned long long)sbd.fssize, sbd.bsize);
 			} else {
 				printf("This is %s metadata\n", sbd.gfs1 ?
 				       "gfs (not gfs2)" : "gfs2");
@@ -948,7 +938,7 @@ static int restore_data(int fd, gzFile *gzin_fd, int printblocksonly,
 			}
 		} else {
 			warm_fuzzy_stuff(savedata->blk, FALSE);
-			if (savedata->blk >= last_fs_block) {
+			if (savedata->blk >= sbd.fssize) {
 				printf("\nOut of space on the destination "
 				       "device; quitting.\n");
 				break;
@@ -973,17 +963,19 @@ static int restore_data(int fd, gzFile *gzin_fd, int printblocksonly,
 				exit(-1);
 			}
 			writes++;
+			if (writes % 1000 == 0)
+				fsync(fd);
 		}
 		blks_saved++;
 	}
 	if (!printblocksonly && !find_highblk)
-		warm_fuzzy_stuff(last_fs_block, TRUE);
+		warm_fuzzy_stuff(sbd.fssize, TRUE);
 	if (find_highblk) {
 		printf("File system size: %lld (0x%llx) blocks, aka %sB\n",
 		       (unsigned long long)highest_valid_block,
 		       (unsigned long long)highest_valid_block,
 		       anthropomorphize(highest_valid_block * sbd.bsize));
-		last_fs_block = highest_valid_block;
+		sbd.fssize = highest_valid_block;
 	}
 	return 0;
 }
